/**
 * dhwk kernel driver
 * Copyright (C) 2006:
 *   Daniel -Blackamp- Michalik
 *   Sebastian tokkee Harl
 *   Florian octo Forster
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 **/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/stat.h> /* access permissions of proc entries */
#include <linux/poll.h>

#include <asm/io.h> /* inw() */
#include <asm/ioctls.h>
#include <asm/system.h>   /* cli(), *_flags */
#include <asm/uaccess.h> /* copy_to_user(), copy_from_user() */

#include "dhwk.h"

MODULE_AUTHOR ("Daniel Michalik, Florian Forster and Sebastian Harl");
MODULE_LICENSE ("GPL");

static dhwk_data_t dhwk_global;

void dhwk_write_hardware (void *ptr);
DECLARE_WORK(dhwk_work_write, dhwk_write_hardware, NULL);

void dhwk_read_hardware (void *ptr);
DECLARE_WORK(dhwk_work_read, dhwk_read_hardware, NULL);

int dhwk_ioctl (struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg);
int dhwk_reset_hw (void);

static void dhwk_print_word (const char *prefix, int word)
{
	printk ("%s0x%04x\n", prefix, word);
}

static void dhwk_print_buffer (const char *prefix, const dhwk_buffer_t *buffer)
{
	unsigned int buf;
	
	printk ("%sbuffer = 0x%p, pos_read = %i, pos_write = %i\n",
			prefix, (void *) buffer, buffer->pos_read, buffer->pos_write);
	buf = inw (DHWK_ADDR_SEND);
	dhwk_print_word (KERN_DEBUG DHWK_MODULE_NAME
			": FLAGS = ", buf & 0xFF00);
}

static int dhwk_proc_recv_read(char *buffer, char **start, off_t offset, int count, int *peof, void *dat)
{
	unsigned int buf = inw(DHWK_ADDR_INTM);
	if(count >= 1)
	{
		if(DHWK_RECV_ENABLED(buf))
		{
			buffer[0] = '1';
		}
		else {
			buffer[0] = '0';
		}
		*peof = 1;
		return 1;
	}
	return 0;
}

static int dhwk_proc_send_read(char *buffer, char **start, off_t offset, int count, int *peof, void *dat)
{
	unsigned int buf = inw(DHWK_ADDR_INTM);
	if(count >= 1)
	{
		if(DHWK_SEND_ENABLED(buf))
		{
			buffer[0] = '1';
		}
		else {
			buffer[0] = '0';
		}
		*peof = 1;
		return 1;
	}
	return 0;
	
}

static int dhwk_proc_info_read(char *buffer, char **start, off_t offset, int count, int *peof, void *dat)
{
	int ret = 0;

	/* Don't use scnprintf for strings longer 75 characters! */
	if (count < 75)
	{
		return 0;
	}
	
	switch (offset)
	{
		case 0:
			ret = scnprintf(buffer, count, "Version: %02i.%02i rev. %i, group %i\n",
					DHWK_DAY, DHWK_MONTH, DHWK_VERSION, DHWK_GROUP);
			break;
		case 1:
			ret = scnprintf(buffer, count, "Authors: Daniel Michalik, Florian Forster and Sebastian Harl\n");
			break;
		case 2:
			ret = scnprintf(buffer, count, "IRQ mask: 0x%02x\n", inw (DHWK_ADDR_INTM) & 0x00ff);
			break;
		case 3:
			ret = scnprintf(buffer, count, "RECV: %s\n", 
					DHWK_RECV_ENABLED (inw (DHWK_ADDR_INTM))
						? "enabled" 
						: "disabled");
			break;
		case 4:
			ret = scnprintf(buffer, count, "SEND: %s\n", 
					DHWK_SEND_ENABLED (inw (DHWK_ADDR_INTM))
						? "enabled" 
						: "disabled");
			break;
		default:
			*peof = 1;
	}
	
	*start = (char *) 1;
	return ret;
}

static int dhwk_proc_recv_write(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char *buf;
	int not_copied;

	buf = kmalloc (count, GFP_KERNEL);
	if (!buf)
	{
		return (-ENOMEM);
	}

	not_copied = copy_from_user (buf, buffer, count);
	if(buf[0] == '0') {
		unsigned int reg;
		reg = inw(DHWK_ADDR_INTM);
		outw(reg & ~DHWK_ENABLE_RECV, DHWK_ADDR_INTM);
	}
	else {
		unsigned int reg;
		reg = inw(DHWK_ADDR_INTM);
		outw(reg | DHWK_ENABLE_RECV, DHWK_ADDR_INTM);
	}

	kfree(buf);
	return count - not_copied;
}

static int dhwk_proc_send_write(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char *buf;
	int not_copied;

	buf = kmalloc (count, GFP_KERNEL);
	if (!buf)
	{
		return (-ENOMEM);
	}

	not_copied = copy_from_user (buf, buffer, count);
	if(buf[0] == '0') {
		unsigned int reg;
		reg = inw(DHWK_ADDR_INTM);
		outw(reg & ~DHWK_ENABLE_SEND, DHWK_ADDR_INTM);
	}
	else {
		unsigned int reg;
		reg = inw(DHWK_ADDR_INTM);
		outw(reg | DHWK_ENABLE_SEND, DHWK_ADDR_INTM);
	}

	kfree(buf);
	return count - not_copied;
}

static void dhwk_tasklet (unsigned long ptr)
{
	int buf;
	int handled = 0;

	buf = inw (DHWK_ADDR_INTR);

	if (buf & DHWK_INTR_SEND_ERROR)
	{
		handled = 1;
		printk (KERN_EMERG DHWK_MODULE_NAME
				": Error interrupt. I'll reset the hardware and you better run!\n");
		/* Reset the hardware */
		dhwk_reset_hw();
		return;
	}

	if ((buf & 0x00FF) == 0x0000)
	{
		handled = 1;
		printk (KERN_WARNING DHWK_MODULE_NAME
				": Caught manual interrupt.\n");
		return;
	}

	if (buf & DHWK_INTR_SOFTIRQ)
	{
		unsigned int tempbuf;
		tempbuf = inw (DHWK_ADDR_INTM);
		outw(tempbuf & ~DHWK_SOFTIRQ, DHWK_ADDR_INTM);

		handled = 1;
		printk (KERN_WARNING DHWK_MODULE_NAME
				": Caught software interrupt.\n");
		return;
	}

	if (buf & DHWK_INTR_RECV_EF_UNSET)
	{
		handled = 1;
		printk (KERN_DEBUG DHWK_MODULE_NAME
				": Caught RECV_EF_UNSET interrupt.\n");
		schedule_work (&dhwk_work_read);
	}

	if (buf & DHWK_INTR_SEND_FF_UNSET)
	{
		handled = 1;
		printk (KERN_DEBUG DHWK_MODULE_NAME
				": Caught SEND_FF_UNSET interrupt.\n");
		schedule_work (&dhwk_work_write);
	}

	if (!handled)
	{
		printk(KERN_DEBUG DHWK_MODULE_NAME
				": Caught unhandled interrupt. WTF?\n");
	}
}

static irqreturn_t dhwk_interrupt (int irq, void *ptr, struct pt_regs *regs)
{
	if (ptr != (void *) &dhwk_global)
		return (IRQ_NONE);

	tasklet_schedule (&dhwk_global.tasklet);
	return (IRQ_HANDLED);
}

int dhwk_ioctl (struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
	char string[42];
	int buf = 0;
	printk(KERN_DEBUG DHWK_MODULE_NAME ": dhwk_ioctl (cmd = %x, arg = %ld)\n", cmd, arg);
	switch(cmd)
	{
		case DHWK_IOCTL_GETGROUP:
			scnprintf(string, 42, "Hardware version: %02i.%02i rev. %i, group %i",
					DHWK_DAY, DHWK_MONTH, DHWK_VERSION, DHWK_GROUP);
			if(copy_to_user((void *) arg, string, 42)) {
				return (-EFAULT);
			}
			break;
		case DHWK_IOCTL_SOFTIRQ:
			buf = inw(DHWK_ADDR_INTM);
			outw(buf | DHWK_SOFTIRQ, DHWK_ADDR_INTM);
			break;
		case DHWK_IOCTL_CHANGE_SEND:
			if(arg)
			{
				buf = inw(DHWK_ADDR_INTM);
				outw(buf | DHWK_ENABLE_SEND, DHWK_ADDR_INTM);
			} 
			else {
				buf = inw(DHWK_ADDR_INTM);
				outw(buf & ~DHWK_ENABLE_SEND, DHWK_ADDR_INTM);
			}
			break;
		case DHWK_IOCTL_CHANGE_RECV:
			if(arg)
			{
				buf = inw(DHWK_ADDR_INTM);
				outw(buf | DHWK_ENABLE_RECV, DHWK_ADDR_INTM);
			} 
			else {
				buf = inw(DHWK_ADDR_INTM);
				outw(buf & ~DHWK_ENABLE_RECV, DHWK_ADDR_INTM);
			}
			break;
		case TCGETS:
			/* EVIL!!!! */
			return (0);
		case DHWK_IOCTL_RESET:
			dhwk_reset_hw();
			printk (KERN_DEBUG DHWK_MODULE_NAME ": dhwk_ioctl: Hardware resetted.\n");
			break;
		default:
			return (-EINVAL);
	}
	return 0;
}

int dhwk_open (struct inode *dev, struct file *fp)
{
	int mode = fp->f_flags & (O_RDONLY | O_WRONLY | O_RDWR);

	printk (KERN_WARNING DHWK_MODULE_NAME ": dhwk_open: mode = 0x%04x\n", mode);

	if ((mode == O_RDWR) || (mode == O_WRONLY))
	{
		if (dhwk_global.open_count_write >= DHWK_MAX_WRITE)
			return -EBUSY;

		dhwk_global.open_count_write++;
	}

	if ((mode == O_RDWR) || (mode == O_RDONLY))
	{
		if (dhwk_global.open_count_read >= DHWK_MAX_READ)
			return -EBUSY;
		
		dhwk_global.open_count_read++;
	}

	printk (KERN_WARNING DHWK_MODULE_NAME ": dhwk_open (%p): "
			"Write count = %i; read count = %i\n",
			(void *) fp,
			dhwk_global.open_count_write,
			dhwk_global.open_count_read);

	return (0);
}

int dhwk_release (struct inode *dev, struct file *fp)
{
	int mode = fp->f_flags & (O_RDONLY | O_WRONLY | O_RDWR);

	if ((mode == O_RDWR) || (mode == O_WRONLY))
		dhwk_global.open_count_write--;

	if ((mode == O_RDWR) || (mode == O_RDONLY))
		dhwk_global.open_count_read--;

	printk (KERN_WARNING DHWK_MODULE_NAME ": dhwk_release (%p): "
			"Write count = %i; read count = %i\n",
			(void *) fp,
			dhwk_global.open_count_write,
			dhwk_global.open_count_read);

	return (0);
}

void dhwk_write_hardware (void *ptr)
{
	spinlock_t *lock;
	unsigned long flags;

	int buf;
	int hw_free;
	int hw_sent;

	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_write_hardware[0]: send_buffer: ", dhwk_global.buffer_send);

	lock = &dhwk_global.buffer_send->lock;

	spin_lock_irqsave (lock, flags);
	hw_free = 0;
	hw_sent = 0;

	while (1)
	{
		if (dhwk_global.buffer_send->pos_read == dhwk_global.buffer_send->pos_write)
			break;

		if (hw_sent >= DHWK_BLOCK_SIZE)
		{
			spin_unlock_irqrestore (lock, flags);
			schedule ();

			spin_lock_irqsave (lock, flags);
			hw_free = 0;
			hw_sent = 0;

			if (dhwk_global.buffer_send->pos_read == dhwk_global.buffer_send->pos_write)
				break;
		}

		if (hw_free == 0)
		{
			buf = inw (DHWK_ADDR_SEND);

			if (DHWK_SEND_EF (buf))
				hw_free = DHWK_BUFFER_FULL_SIZE;
			else if (!DHWK_SEND_HF (buf))
				hw_free = DHWK_BUFFER_HALF_SIZE;
			else if (!DHWK_SEND_FF (buf))
				hw_free = 1;
			else
			{
				break;
			}

			if (hw_free > DHWK_BLOCK_SIZE)
				hw_free = DHWK_BLOCK_SIZE;
		}

		outw ((int) dhwk_global.buffer_send->data[dhwk_global.buffer_send->pos_read], DHWK_ADDR_SEND);

		dhwk_global.buffer_send->pos_read = (dhwk_global.buffer_send->pos_read + 1) % DHWK_BUFFER_SIZE;
		hw_free--;
		hw_sent++;
	}

	spin_unlock_irqrestore (lock, flags);
	
	/* Awake waiting processes */
	dhwk_global.buffer_send->ready = 1;
	wake_up_interruptible (&dhwk_global.buffer_send->wait);
	
	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_write_hardware[1]: send_buffer: ", dhwk_global.buffer_send);
	return;
}

ssize_t dhwk_write (struct file *fp, const char *buf, size_t size, loff_t *pos)
{
	int retval = 0;
	spinlock_t *lock;
	unsigned long flags;

	int   buffer_used;
	int   buffer_free;
	char *buffer_ptr;

	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_write[0]: send_buffer: ", dhwk_global.buffer_send);

	lock = &dhwk_global.buffer_send->lock;

	/* Wait for free space in the buffer */
	while (1)
	{
		spin_lock_irqsave (lock, flags);

		buffer_used = (dhwk_global.buffer_send->pos_write - dhwk_global.buffer_send->pos_read + DHWK_BUFFER_SIZE) % DHWK_BUFFER_SIZE;
		buffer_free = DHWK_BUFFER_SIZE - (buffer_used + 1);
		buffer_ptr  = dhwk_global.buffer_send->data + dhwk_global.buffer_send->pos_write;

		if (buffer_free != 0)
			break;

		spin_unlock_irqrestore (lock, flags);

		if (DHWK_IS_NONBLOCK (fp))
		{
			return (-EAGAIN);
		}

		dhwk_global.buffer_send->ready = 0;
		if (0 != wait_event_interruptible (dhwk_global.buffer_send->wait, dhwk_global.buffer_send->ready))
		{
			printk (KERN_DEBUG DHWK_MODULE_NAME
					": dhwk_write: Interrupted by signal.\n");
			return (-EINTR);
		}
	}

	/* `buffer_free' is now garuanteed to be non-zero */

	if (size > buffer_free)
		size = buffer_free;

	/* Write till the end first */
	if ((dhwk_global.buffer_send->pos_write + size) > DHWK_BUFFER_SIZE)
	{
		int bytes_first;
		int bytes_second;


		/*
		 *           pos_write ---.
		 *                        V
		 * +--------------+- - - -+-------------+
		 * ! bytes_second !  ...  ! bytes_first !
		 * +--------------+- - - -+-------------+
		 */
		bytes_first  = DHWK_BUFFER_SIZE - dhwk_global.buffer_send->pos_write;
		bytes_second = size - bytes_first;

		if (copy_from_user (buffer_ptr, buf, bytes_first))
		{
			spin_unlock_irqrestore (lock, flags);
			return (-EFAULT);
		}

		buffer_ptr = dhwk_global.buffer_send->data;
		buf += bytes_first;
		size = bytes_second;
		retval += bytes_first;

		dhwk_global.buffer_send->pos_write = 0;
	}

	/* Copy the rest. It's garantueed to fit in the free space now.. */
	if (copy_from_user (buffer_ptr, buf, size))
	{
		spin_unlock_irqrestore (lock, flags);
		return (-EFAULT);
	}

	dhwk_global.buffer_send->pos_write += size;
	dhwk_global.buffer_send->pos_write %= DHWK_BUFFER_SIZE;

	spin_unlock_irqrestore (lock, flags);

	retval += size;

	if (DHWK_IS_NONBLOCK (fp))
		schedule_work (&dhwk_work_write);
	else
		dhwk_write_hardware (NULL);

	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_write[1]: send_buffer: ", dhwk_global.buffer_send);
	printk (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_write: return (%i);\n", retval);
	return (retval);
}

void dhwk_read_hardware (void *ptr)
{
	unsigned int recv_buf;
	spinlock_t *lock = &dhwk_global.buffer_recv->lock;

	unsigned long flags;

	int pos;
	
	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_read_hardware[0]: recv_buffer: ", dhwk_global.buffer_recv);

	spin_lock_irqsave (lock, flags);
	
	while (((dhwk_global.buffer_recv->pos_write + 1) % DHWK_BUFFER_SIZE)
			!= dhwk_global.buffer_recv->pos_read)
	{
		recv_buf = inw (DHWK_ADDR_RECV);

		if (DHWK_RECV_EF (recv_buf))
		{
			break;
		}

		pos = dhwk_global.buffer_recv->pos_write;
		dhwk_global.buffer_recv->data[pos] = (char) (recv_buf & 0x00FF);
		dhwk_global.buffer_recv->pos_write = (pos + 1) % DHWK_BUFFER_SIZE;
	}
	
	spin_unlock_irqrestore (lock, flags);

	dhwk_global.buffer_recv->ready = 1;
	wake_up_interruptible (&dhwk_global.buffer_recv->wait);
	
	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_read_hardware[1]: recv_buffer: ", dhwk_global.buffer_recv);
	return;
}

ssize_t dhwk_read (struct file *fp, char *buf, size_t size, loff_t *pos)
{
	spinlock_t *lock;

	int len_fill;
	unsigned long flags;

	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_read[0]: recv_buffer: ", dhwk_global.buffer_recv);

	lock = &dhwk_global.buffer_recv->lock;

	/* should work without irqsave, too */
	spin_lock_irqsave (lock, flags);
	while (1)
	{
		unsigned int buf;
		
		/* buffer is not empty */
		if (dhwk_global.buffer_recv->pos_write != dhwk_global.buffer_recv->pos_read)
			break;

		/* buffer is empty; check if we have data in the fifo */
		buf = inw (DHWK_ADDR_SEND);
		if (!DHWK_RECV_EF (buf))
		{
			if (DHWK_IS_NONBLOCK (fp))
			{
				schedule_work (&dhwk_work_read);
				return (-EAGAIN);
			}
			else {
				dhwk_read_hardware (NULL);
				continue;
			}
		}
		
		/* we only reach this section if fifo and buffer are empty */
		if (DHWK_IS_NONBLOCK (fp))
		{
			spin_unlock_irqrestore (lock, flags);
			return (-EAGAIN);
		}

		/* Go to sleep until the wait-queue wakes us */
		spin_unlock_irqrestore (lock, flags);

		dhwk_global.buffer_recv->ready = 0;
		if (0 != wait_event_interruptible (dhwk_global.buffer_recv->wait, dhwk_global.buffer_recv->ready))
		{
			printk (KERN_DEBUG DHWK_MODULE_NAME
					": dhwk_read: Interrupted by signal.\n");
			return (-EINTR);
		}
		
		spin_lock_irqsave (lock, flags);
	}

	len_fill = (dhwk_global.buffer_recv->pos_write - dhwk_global.buffer_recv->pos_read + DHWK_BUFFER_SIZE) % DHWK_BUFFER_SIZE;

	/* We will copy `size' bytes to the user.
	 * 0 < size <= len_fill <= DHWK_BUFFER_SIZE */
	if (size > len_fill)
		size = len_fill;

	/* Don't wrap around at the end of the buffer. This makes stuff a lot easier.. */
	if ((dhwk_global.buffer_recv->pos_read + size) > DHWK_BUFFER_SIZE)
		size = DHWK_BUFFER_SIZE - dhwk_global.buffer_recv->pos_read;

	if (copy_to_user (buf, dhwk_global.buffer_recv->data + dhwk_global.buffer_recv->pos_read, size) != 0)
	{
		spin_unlock_irqrestore (lock, flags);
		return (-EFAULT);
	}

	dhwk_global.buffer_recv->pos_read = (dhwk_global.buffer_recv->pos_read + size) % DHWK_BUFFER_SIZE;

	spin_unlock_irqrestore (lock, flags);

	dhwk_print_buffer (KERN_DEBUG DHWK_MODULE_NAME 
			": dhwk_read[1]: recv_buffer: ", dhwk_global.buffer_recv);
	printk (KERN_DEBUG DHWK_MODULE_NAME
			": dhwk_read: return (%i);\n", size);
	return (size);
}

static unsigned int dhwk_poll(struct file *fp, poll_table *pt)
{
	unsigned int mask = 0;

	int len_read;
	int len_write;

	poll_wait(fp, &dhwk_global.buffer_recv->wait, pt);
	poll_wait(fp, &dhwk_global.buffer_send->wait, pt);

	len_read = (dhwk_global.buffer_recv->pos_write - dhwk_global.buffer_recv->pos_read + DHWK_BUFFER_SIZE) % DHWK_BUFFER_SIZE;
	len_write = (dhwk_global.buffer_send->pos_write - dhwk_global.buffer_send->pos_read + DHWK_BUFFER_SIZE) % DHWK_BUFFER_SIZE;
	
	if (len_read > 0)
		mask |= POLLIN | POLLRDNORM;
	if ((len_write + 1) < DHWK_BUFFER_SIZE)
		mask |= POLLOUT | POLLWRNORM;

	return mask;
}

static struct file_operations fops = {
	.read    = dhwk_read,
	.write   = dhwk_write,
	.ioctl   = dhwk_ioctl,
	.open    = dhwk_open,
	.release = dhwk_release,
	.poll    = dhwk_poll
};

static int dhwk_irq_probing (void)
{
	unsigned long irqs;
	int ret;

	sti ();
	irqs = probe_irq_on();

	/* Generate an interrupt */
	outw(DHWK_SOFTIRQ, DHWK_ADDR_INTM);
	outw(0x0000, DHWK_ADDR_INTM);

	ret = probe_irq_off (irqs);

	if (ret <= 0)
	{
		printk (KERN_WARNING DHWK_MODULE_NAME
				": dhwk_irq_probing: Could not find interrupt: status = %i\n",
				ret);
		return (-1);
	}

	return (ret);
}

static int dhwk_check_magic (void)
{
	int magic = 0;

	magic = inw (DHWK_ADDR_MGIC);

	/* parens to make compiler happy */
	if (magic != (DHWK_MAGIC))
	{
		printk (KERN_WARNING DHWK_MODULE_NAME
				": Hardware returned wrong magic: Expected `0x%4x', got `0x%4x'\n",
				DHWK_MAGIC, magic);
		return (-1);
	}

	printk (KERN_INFO DHWK_MODULE_NAME
			": Hardware version: %02i.%02i rev. %i, group %i\n",
			DHWK_DAY, DHWK_MONTH, DHWK_VERSION, DHWK_GROUP);

	return (0);
}

static dhwk_buffer_t *dhwk_init_buffer (void)
{
	dhwk_buffer_t *ret;

	ret = (dhwk_buffer_t *) kmalloc (sizeof (dhwk_buffer_t), GFP_KERNEL);
	spin_lock_init (&ret->lock);
	init_waitqueue_head (&ret->wait);
	ret->pos_read  = 1;
	ret->pos_write = 1;

	return (ret);
}

static int dhwk_reset_hw (void)
{
	outw (DHWK_RESET, DHWK_ADDR_INTM);
	outw (DHWK_INTR_SEND_ERROR
			| DHWK_INTR_RECV_EF_UNSET
			/* | DHWK_INTR_RECV_HF_ISSET */
			/* | DHWK_INTR_RECV_FF_ISSET */
			| DHWK_INTR_SOFTIRQ
			/* | DHWK_INTR_SEND_EF_ISSET */
			/* | DHWK_INTR_SEND_HF_UNSET */
			| DHWK_INTR_SEND_FF_UNSET
			| DHWK_ENABLE_RECV
			| DHWK_ENABLE_SEND,
			DHWK_ADDR_INTM);
}

static int __init dhwk_init (void)
{
	int status;
	unsigned short buf;

	if ((status = dhwk_check_magic ()) != 0)
		return (status);

	if ((dhwk_global.interrupt = dhwk_irq_probing ()) <= 0)
	{
		printk (KERN_WARNING DHWK_MODULE_NAME
				": dhwk_init: Could not find interrupt: status = %i\n",
				dhwk_global.interrupt);
		return (-1);
	}

	dhwk_global.buffer_recv = dhwk_init_buffer ();
	dhwk_global.buffer_send = dhwk_init_buffer ();

	tasklet_init (&dhwk_global.tasklet, dhwk_tasklet, 0L);

	if ((status = register_chrdev (DHWK_MAJOR, DHWK_MODULE_NAME, &fops)) < 0)
	{
		printk (KERN_WARNING DHWK_MODULE_NAME
				": Can't get major %i. status = %i\n",
				DHWK_MAJOR, status);
		return (status);
	}

	if (DHWK_MAJOR != 0)
		dhwk_global.major = DHWK_MAJOR;
	else
		dhwk_global.major = status;

	printk (KERN_INFO DHWK_MODULE_NAME
			": Registered character device with major %i\n",
			dhwk_global.major);

	dhwk_reset_hw();


	if ((status = request_irq (dhwk_global.interrupt, dhwk_interrupt,
				SA_INTERRUPT | SA_SHIRQ,
				"DHWK", (void *) &dhwk_global)) != 0)
	{
		printk (KERN_WARNING DHWK_MODULE_NAME
				": Cannot register interrupt %i: status = %i\n",
				dhwk_global.interrupt, status);
		
		if ((status = unregister_chrdev (dhwk_global.major, DHWK_MODULE_NAME)) < 0)
		{
			printk (KERN_EMERG DHWK_MODULE_NAME
					": Oh boy, you're in trouble! `unregister_chrdev' returned %i\n",
					status);
			return (-1);
		}

		printk (KERN_INFO DHWK_MODULE_NAME
				": Unregistered character device with major %i\n",
				dhwk_global.major);
		return (-1);
	}

	buf = inw (DHWK_ADDR_INTM);

	printk (KERN_INFO DHWK_MODULE_NAME
			": Registered interrupt %i with IRQ-mask 0x%02x\n",
			dhwk_global.interrupt, buf & 0x00FF);

	dhwk_global.proc_dir = proc_mkdir ("dhwk3", NULL);
	if (dhwk_global.proc_dir) {
		dhwk_global.proc_recv = create_proc_entry ("recv", S_IRUGO | S_IWUSR, dhwk_global.proc_dir);
		if(dhwk_global.proc_recv) {
			dhwk_global.proc_recv->read_proc = dhwk_proc_recv_read;
			dhwk_global.proc_recv->write_proc = dhwk_proc_recv_write;
			dhwk_global.proc_recv->data = NULL;
		}

		dhwk_global.proc_send = create_proc_entry ("send", S_IRUGO | S_IWUSR, dhwk_global.proc_dir);
		if(dhwk_global.proc_send) {
			dhwk_global.proc_send->read_proc = dhwk_proc_send_read;
			dhwk_global.proc_send->write_proc = dhwk_proc_send_write;
			dhwk_global.proc_send->data = NULL;
		}

		dhwk_global.proc_info = create_proc_entry ("info", S_IRUGO, dhwk_global.proc_dir);
		if(dhwk_global.proc_info) {
			dhwk_global.proc_info->read_proc = dhwk_proc_info_read;
			dhwk_global.proc_info->data = NULL;
		}
	}
	else {
		 printk (KERN_DEBUG DHWK_MODULE_NAME
				 ": Couldn't create proc directory\n");
	}
	return (0); /* success */
}

static void __exit dhwk_cleanup (void)
{
	int status;
	
	kfree (dhwk_global.buffer_recv);
	kfree (dhwk_global.buffer_send);

	/* Remove proc entries */
	if (dhwk_global.proc_info)
		remove_proc_entry("info", dhwk_global.proc_dir);

	if (dhwk_global.proc_send)
		remove_proc_entry("send", dhwk_global.proc_dir);

	if (dhwk_global.proc_recv)
		remove_proc_entry("recv", dhwk_global.proc_dir);

	if (dhwk_global.proc_dir)
		remove_proc_entry("dhwk3", NULL);

	/* Disable all interrupts */
	outw (0x0000, DHWK_ADDR_INTM);

	free_irq (dhwk_global.interrupt, (void *) &dhwk_global);

	printk (KERN_INFO DHWK_MODULE_NAME
			": Unregistered interrupt %i\n",
			dhwk_global.interrupt);

	if ((status = unregister_chrdev (dhwk_global.major, DHWK_MODULE_NAME)) < 0)
	{
		printk (KERN_EMERG DHWK_MODULE_NAME
				": Oh boy, you're in trouble! `unregister_chrdev' returned %i\n",
				status);
		return;
	}

	printk (KERN_INFO DHWK_MODULE_NAME
			": Unregistered character device with major %i\n",
			dhwk_global.major);

	printk (KERN_INFO DHWK_MODULE_NAME
			": Module removed\n");
	return;
}

module_init (dhwk_init);
module_exit (dhwk_cleanup);
